from tktl.commands.health import GetGrpcHealthCommand, GetRestHealthCommand


def test_health(logged_in_context, test_user_deployed_repos):
    for repo, branch, endpoint in test_user_deployed_repos[2:]:
        cmd = GetRestHealthCommand(
            repository=repo, branch_name=branch, local=False, skip_auth=True
        )
        result = cmd.execute()
        assert result.status_code == 204

    grpc_repo_name, grpc_repo_branch, grpc_repo_endpoint = test_user_deployed_repos[-1]
    cmd = GetGrpcHealthCommand(
        repository=grpc_repo_name,
        branch_name=grpc_repo_branch,
        local=False,
        skip_auth=True,
    )
    result = cmd.execute()
    assert result is True



from collections.abc import Iterable, Sequence
from functools import singledispatch
from typing import (  # type: ignore
    Any,
    Dict,
    List,
    Optional,
    Type,
    Union,
    _GenericAlias,
    _SpecialForm,
)

import numpy  # type: ignore
import pandas  # type: ignore
from pandas.core.dtypes.inference import is_sequence  # type: ignore
from pydantic import create_model
from pydantic.main import ModelMetaclass

from tktl.core.exceptions import TaktileSdkError
from tktl.core.serializers.base import CustomDeserializingModelT
from tktl.core.serializers.rest.schema import (
    get_array_model,
    get_custom_model,
    get_dataframe_model,
    get_flat_array_model,
    get_mapping_model,
    get_nested_sequence_model,
    get_series_model,
    get_single_value_model,
)
from tktl.core.serializers.rest.utils import pandas_to_python_type
from tktl.core.t import RestSchemaTypes


@singledispatch
def to_pydantic(
    value: Any, name: str = RestSchemaTypes.SINGLE_VALUE.value, unique_id: str = None
) -> Union[Type[CustomDeserializingModelT], Type[_GenericAlias], _GenericAlias]:
    """
    Single dispatch function for serializing values of each of the supported kinds. Each
    of the dispatched functions' implementation is tailored for one of Taktile's supported
    input types. All the implementations return a custom Pydantic model that implements
    a `deserialize` method to be used when deserializing incoming data as a REST request

    Parameters
    ----------
    value : Any
        Input value to be serialized
    name : str
        Name of the model
    unique_id : str
        Unique identifier of the model

    Returns
    -------
    single_value_model : CustomDeserializingModelT
        The Pydantic-based model
    """
    single_value_model = get_single_value_model(value)
    single_value_model.__name__ = get_model_unique_name(name=name, unique_id=unique_id)
    return single_value_model


@to_pydantic.register
def _custom(
    value: ModelMetaclass,
    name: str = RestSchemaTypes.CUSTOM_MODEL.value,
    unique_id: str = None,
):
    model = get_custom_model(value)
    model.__name__ = get_model_unique_name(name=name, unique_id=unique_id)
    return model


@to_pydantic.register
def _df(
    value: pandas.DataFrame,
    name: str = RestSchemaTypes.DATAFRAME.value,
    unique_id: str = None,
    nullable=True,
):
    type_map = {}
    example_map = {}

    for col in value.columns:
        var_type = pandas_to_python_type(value[col])
        if var_type is None:
            types = (type(v) for _, v in value[col].iteritems() if not pandas.isna(v))
            try:
                var_type = next(types)  # type of first item that is not None
            except StopIteration:
                raise TaktileSdkError(
                    f"Column {col} contains all null values and unable to infer its type with .dtype"
                )

        if nullable:
            var_type = Optional[var_type]

        type_map[col] = (var_type, ...)

        # Make nxt_val (pandas.isna is recursive so need to check for iterable first)
        vals = (
            v
            for _, v in value[col].iteritems()
            if isinstance(v, Iterable) or not pandas.isna(v)
        )

        try:
            nxt_val = next(vals)
        except StopIteration:
            nxt_val = None  # python built-in version of NaN

        example_map[col] = nxt_val

    index_type = type(next((t for t in value.index)))
    type_map["index"] = (index_type, None)  # type: ignore

    base_model = create_model(name, **type_map)  # type: ignore
    inner_model_unique_id = get_model_unique_name(
        name, unique_id=f"{unique_id}__dt_validate"
    )
    dataframe_model = get_dataframe_model(
        base_model=base_model, example=[example_map], unique_id=inner_model_unique_id
    )
    dataframe_model.__name__ = get_model_unique_name(name=name, unique_id=unique_id)
    return dataframe_model


@to_pydantic.register
def _series(
    value: pandas.Series,
    name: str = RestSchemaTypes.SERIES.value,
    unique_id: str = None,
) -> Type[CustomDeserializingModelT]:
    nullable = True if value.isna().sum() > 0 else False
    type_name = "Outcome" if not value.name else value.name
    type_map = {}

    try:
        nxt_val = next(
            v
            for v in value.to_dict().values()
            if isinstance(v, Iterable) or not pandas.isna(v)
        )
    except StopIteration:
        nxt_val = None

    example_map = {type_name: [nxt_val]}

    types = (type(v) for k, v in value.to_dict().items() if not pandas.isna(v))
    var_type = next(types)  # type of first item that is not None
    if var_type == bool:
        var_type = Union[var_type, float]  # type: ignore
    if nullable:
        var_type = Optional[var_type]  # type: ignore
    type_map[type_name] = (List[var_type], None)  # type: ignore
    base_model = create_model(name, **type_map)  # type: ignore
    series_model = get_series_model(
        series=value, base_model=base_model, example=example_map
    )
    series_model.__name__ = get_model_unique_name(name=name, unique_id=unique_id)
    return series_model


@to_pydantic.register
def _array(
    value: numpy.ndarray, name: str = RestSchemaTypes.ARRAY.value, unique_id: str = None
) -> Type[CustomDeserializingModelT]:
    as_list = value.tolist()
    if len(value.shape) == 1:
        first_non_empty = next(v for v in as_list if not pandas.isna(v))
        array_model = get_flat_array_model(
            type(first_non_empty), example=[first_non_empty]
        )
        name = RestSchemaTypes.FLAT_ARRAY.value
        array_model.__name__ = name if not unique_id else f"{name}__{unique_id}"
    else:
        example = replace_nans_and_infs(value[0]).tolist()
        list_model = get_nested_sequence_model(as_list)
        array_model = get_array_model(list_model, example=[example])
        array_model.__name__ = get_model_unique_name(name=name, unique_id=unique_id)
    return array_model


@to_pydantic.register
def _sequence(
    value: Sequence, name: str = RestSchemaTypes.SEQUENCE.value, unique_id: str = None
) -> Type[CustomDeserializingModelT]:
    if isinstance(value, list) or isinstance(value, tuple):
        if len(value) > 0:
            first = value[0]
            if isinstance(first, list):
                sequence_model = get_nested_sequence_model(first)
                sequence_model.__name__ = get_model_unique_name(
                    name=name, unique_id=unique_id
                )

            elif isinstance(first, dict):
                mapping_model = _dict_model(
                    value=first, name=RestSchemaTypes.DICT.value
                )
                sequence_model = get_mapping_model(base_model=mapping_model)
                sequence_model.__name__ = get_model_unique_name(
                    name=name, unique_id=unique_id
                )
            # more specific than numpy's is_sequence implementation
            elif not is_sequence(first):
                sequence_model = get_single_value_model(first)
                sequence_model.__name__ = get_model_unique_name(
                    name=name, unique_id=unique_id
                )
            else:
                raise
        else:
            raise ValueError("Input must be non-empty")
    else:
        raise
    return sequence_model


@to_pydantic.register
def _dict(
    value: dict, name: str = RestSchemaTypes.DICT.value, unique_id: str = None
) -> Type[CustomDeserializingModelT]:
    mapping_model = _dict_model(value=value, name=name)
    dict_model = get_mapping_model(mapping_model)
    dict_model.__name__ = get_model_unique_name(name=name, unique_id=unique_id)
    return dict_model


@to_pydantic.register
def _generic(
    value: _GenericAlias, name: str = RestSchemaTypes.ANY.value, unique_id: str = None,
) -> Type[_GenericAlias]:
    return value


@to_pydantic.register
def _any(
    value: _SpecialForm, name: str = RestSchemaTypes.ANY.value, unique_id: str = None,
) -> _GenericAlias:
    return Union[Dict, List, Any]


def _dict_model(value, name):
    types = {k: (type(v), v) for k, v in value.items()}
    return create_model(name, **types)


def _get_df_example(example_map):
    class Config:
        schema_extra = {"example": example_map}

    return Config


def get_model_unique_name(name: str, unique_id: Optional[str]) -> str:
    return name if not unique_id else f"{name}__{unique_id}"


def replace_nans_and_infs(array) -> numpy.ndarray:
    as_float = numpy.isnan(array.astype(numpy.float64))
    as_inf = numpy.isinf(array.astype(numpy.float64))
    replaced_nans = numpy.where(as_float, None, array)
    return numpy.where(as_inf, 0.0, replaced_nans)



import json
from io import StringIO

import yaml
from click.testing import CliRunner

from tktl import main
from tktl.cli import get


def test_get(logged_in_context):
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(main.get_commands)
    assert result.exit_code == 0
    assert "deployments   Get deployment resources\n" in result.output
    assert "repositories  Get repository resources\n" in result.output
    assert "endpoints     Get endpoint resources\n" in result.output


def test_get_deployments(logged_in_context):
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(get.get_deployment_by_repo_id)
    assert result.exit_code == 0
    assert "2732fdb2-67e4-428f-ac7d-72f379376b67" in result.output
    assert "582132db-6687-4c87-9907-ffbc8f03fa07" in result.output

    as_io = StringIO(result.output)
    lines = as_io.readlines()
    assert len(lines) >= 6


def test_get_deployments_with_options(logged_in_context):
    runner = CliRunner()
    result_with_id = runner.invoke(
        get.get_deployment_by_repo_id, "99ffb2dc-b06d-4262-b2e0-34cd40a0c56d"
    )
    assert "refs/heads/master @ 344e3bc" in result_with_id.output
    as_io = StringIO(result_with_id.output)
    lines = as_io.readlines()
    assert len(lines) == 3


def test_json_resources(logged_in_context):
    runner = CliRunner()
    for command in [
        get.get_deployment_by_repo_id,
        get.get_repositories,
        get.get_endpoint_by_deployment_id,
    ]:
        result = runner.invoke(command, ["-O", "json", "-a"])
        assert result.exit_code == 0
        as_io = StringIO(result.output)
        lines = as_io.read()
        loaded = json.loads(lines)
        assert isinstance(loaded, list)
        for item in loaded:
            assert "id" in item.keys() or "deployment_id" in item.keys()


def test_yaml_resources(logged_in_context):
    runner = CliRunner()
    for command in [
        get.get_deployment_by_repo_id,
        get.get_repositories,
        get.get_endpoint_by_deployment_id,
    ]:
        result = runner.invoke(command, ["-O", "yaml", "-a"])
        assert result.exit_code == 0
        as_io = StringIO(result.output)
        lines = as_io.read()
        loaded = [line for line in yaml.safe_load_all(lines)]
        assert len(loaded) >= 3
        assert all([("id" in k.keys() or "deployment_id" in k.keys()) for k in loaded])



import joblib
import pandas as pd
from sklearn.metrics import accuracy_score, roc_auc_score

model = joblib.load("assets/model.joblib")
loans = pd.read_parquet("assets/loans_test.pqt")
label = "Repaid"
X = loans.drop(columns=label)
y = loans[label]
pred = model.predict(X)
probs = model.predict_proba(X)[:, 1]


def test_auc(json_metadata):
    auc = roc_auc_score(y, probs)

    json_metadata["section"] = "Accuracy"
    json_metadata["pass_message"] = f"AUC is acceptable ({auc:.2f})"
    json_metadata["fail_message"] = f"AUC is inacceptable ({auc:.2f})"

    assert 0.6 < auc < 0.7


def test_accuracy(json_metadata):
    accuracy = accuracy_score(y, pred)

    json_metadata["section"] = "Accuracy"
    json_metadata["pass_message"] = f"Accuracy is acceptable ({accuracy:.2f})"
    json_metadata["fail_message"] = f"Accuracy is inacceptable ({accuracy:.2f})"

    assert 0.8 < accuracy < 0.9



import sys
import xmlrpc.client as xmlrpclib
from distutils.version import StrictVersion
from platform import system

from tktl import __version__
from tktl.core.loggers import LOG


class PackageNotFoundError(Exception):
    pass


class VersionChecker(object):
    def is_up_to_date(self, module_name, current_version):
        version_in_repository = self.get_version_from_repository(module_name)

        up_to_date = StrictVersion(current_version) >= StrictVersion(
            version_in_repository
        )
        return up_to_date, version_in_repository

    @staticmethod
    def get_version_from_repository(
        module_name, repository_url="http://pypi.python.org/pypi"
    ):
        pypi = xmlrpclib.ServerProxy(repository_url)
        versions = pypi.package_releases(module_name)
        if not versions:
            raise PackageNotFoundError("Package {} not found".format(module_name))

        return versions[0]


class TaktileVersionChecker(object):
    @classmethod
    def look_for_new_version_with_timeout(cls):
        if not cls._should_check_version():
            return

        if not system() == "Linux":
            cls.look_for_new_version()
            return

        import signal

        class TimeoutError(Exception):
            pass

        def handler(signum, frame):
            raise TimeoutError

        signal.signal(signal.SIGALRM, handler)
        signal.alarm(1)

        try:
            cls.look_for_new_version()
        except TimeoutError:
            pass

        signal.alarm(0)

    @staticmethod
    def look_for_new_version():
        vc = VersionChecker()
        try:
            up_to_date, version_from_repository = vc.is_up_to_date(
                "taktile-cli", __version__
            )
        except Exception as e:
            LOG.error(e)
            return

        if not up_to_date:
            msg = f"""
Warning: this version of the Taktile CLI ({__version__}) is out of date.
Some functionality might not be supported until you upgrade.
Latest available version: {version_from_repository}\n
Run `pip install -U taktile-cli` to upgrade.
"""
            LOG.warning(msg)

    @staticmethod
    def _should_check_version():
        if not hasattr(sys.stdin, "isatty"):
            return False
        if not sys.stdin.isatty() or not sys.stdout.isatty():
            return False

        return True



from typing import List, Optional, Union

from pydantic import BaseModel

from tktl.core.t import InstanceTypeT, ProjectAssetSourceT, ProjectAssetT


class ProjectContentsBase(BaseModel):
    type: str
    name: str
    path: str


class ProjectFile(ProjectContentsBase):
    type: str = "file"


class ProjectFileWithContent(ProjectContentsBase):
    type: str = "file"
    content: str


class ProjectDirectory(ProjectContentsBase):
    type: str = "dir"


class ProjectAsset(ProjectContentsBase):
    calculated_sha: str
    kind: ProjectAssetT
    source: ProjectAssetSourceT
    requires_download: bool


class RestServiceConfigSchema(BaseModel):
    replicas: int
    max_replicas: int
    instance_type: InstanceTypeT


class ArrowServiceConfigSchema(BaseModel):
    replicas: int
    instance_type: InstanceTypeT


class TktlServiceConfigSchema(BaseModel):
    rest: RestServiceConfigSchema
    arrow: ArrowServiceConfigSchema


class TktlYamlConfigSchema(BaseModel):

    """
    deployment_prefix?: master
    service:

      rest:
        instance_type: gp.small
        replicas: 1
        max_replicas: 1

      arrow:
        instance_type: gp.small
        replicas: 1

    version: version
    """

    deployment_prefix: Optional[str]
    service: TktlServiceConfigSchema
    version: str


class TktlYamlConfigValidationError(BaseModel):
    loc: str
    msg: str


class ProjectValidationOutput(BaseModel):
    title: str
    summary: str
    text: str


ProjectContentSingleItemT = Union[ProjectFile, ProjectDirectory]
ProjectContentMultiItemT = List[ProjectContentSingleItemT]
ProjectContentT = Union[
    ProjectContentSingleItemT, ProjectContentMultiItemT, ProjectFileWithContent
]



from typing import List

from beautifultable import BeautifulTable  # type: ignore

from tktl.core.schemas.repository import Endpoint


def get_table_from_list_of_schemas(
    models: List[Endpoint],
    keys: List[str] = None,
    names: List[str] = None,
    maxwidth=140,
):
    if not names:
        names = []
    table = BeautifulTable(
        maxwidth=maxwidth, default_alignment=BeautifulTable.ALIGN_LEFT
    )
    rows = len(models)
    if rows < 1:
        raise ValueError("Nothing to print")

    first = models[0]
    as_dict = first.table_repr(subset=keys)
    if names:
        table.columns.header = list(as_dict.keys())

    if not names and keys:
        table.columns.header = keys
    elif not names and not keys:
        keys = list(first.table_repr().keys())
        table.columns.header = keys
    for i in range(rows):
        if keys:
            table.rows.append([models[i].table_repr()[attr] for attr in keys])
    table.set_style(BeautifulTable.STYLE_NONE)
    return table



import abc
import typing as t

import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from fastapi import Response
from pydantic import BaseModel

from tktl.core.future.t import EndpointKinds, ProfileKinds


class Endpoint(abc.ABC):
    kind: EndpointKinds

    def __init__(
        self,
        name: str,
        func: t.Callable[..., t.Coroutine[t.Any, t.Any, Response]],
        X: t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None] = None,
        y: t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None] = None,
        **kwargs,
    ):
        self._kwargs = kwargs
        self._name = name
        self._func = func
        self._X = X
        self._y = y

    @property
    def kwargs(self):
        return self._kwargs

    @property
    def name(self):
        return self._name

    @property
    def func(self):
        return self._func

    @property
    def X(self):
        return self._X

    @property
    def y(self):
        return self._y

    @staticmethod
    @abc.abstractmethod
    def supported(
        *,
        X: t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None] = None,
        y: t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None] = None,
        profile: t.Optional[str] = None,
    ) -> bool:
        """supported.
        Given the parameters, is this tyep of endpoint supported?

        Parameters
        ----------
        X : t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None]
        y : t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None]
        profile : t.Optional[str]
            the paramters

        Returns
        -------
        bool - true if supported, false else

        """


class BasicEndpoint(Endpoint):
    kind: EndpointKinds = EndpointKinds.BASIC

    @staticmethod
    def supported(
        *,
        X: t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None] = None,
        y: t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None] = None,
        profile: t.Optional[str] = None,
    ) -> bool:
        return profile is None and X is None and y is None


class TypedEndpoint(Endpoint):
    kind: EndpointKinds = EndpointKinds.TYPED

    @staticmethod
    def supported(
        *,
        X: t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None] = None,
        y: t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None] = None,
        profile: t.Optional[str] = None,
    ) -> bool:
        return (
            profile is None
            and isinstance(X, type(BaseModel))
            and isinstance(y, type(BaseModel))
        )


class ArrowEndpoint(TypedEndpoint):
    kind: EndpointKinds = EndpointKinds.ARROW

    @staticmethod
    def supported(
        *,
        X: t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None] = None,
        y: t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None] = None,
        profile: t.Optional[str] = None,
    ) -> bool:
        return (
            profile is None
            and isinstance(X, (np.ndarray, pd.DataFrame, pd.Series))
            and isinstance(y, (np.ndarray, pd.DataFrame, pd.Series))
        )


class ProfiledEndpoint(ArrowEndpoint):
    kind: EndpointKinds = EndpointKinds.PROFILED

    @staticmethod
    def supported(
        *,
        X: t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None] = None,
        y: t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None] = None,
        profile: t.Optional[str] = None,
    ) -> bool:
        return (
            profile in ProfileKinds.set()
            and isinstance(X, (np.ndarray, pd.DataFrame, pd.Series))
            and isinstance(y, (np.ndarray, pd.DataFrame, pd.Series))
        )



import click

from tktl.commands import login as login_commands


@click.command("login", help="Log in & store api key")
@click.argument("api_key", required=True)
def login(api_key: str):
    command = login_commands.LogInCommand()
    command.execute(api_key=api_key)


@click.command("logout", help="Log out & remove apiKey from config file")
def logout():
    command = login_commands.LogOutCommand()
    command.execute()



import os

import pytest
import requests
from pydantic import BaseModel

from tktl.core.clients.http_client import API, interpret_response
from tktl.core.config import settings
from tktl.core.exceptions.exceptions import APIClientException
from tktl.core.managers.auth import AuthConfigManager


def test_instantiate_api_client():
    key = os.environ["TEST_USER_API_KEY"]
    AuthConfigManager.set_api_key(key)
    client = API(api_url=settings.DEPLOYMENT_API_URL)
    assert "X-Api-Key" in client.default_headers
    assert client.default_headers["X-Api-Key"] == key
    assert (
        client.get_path(f"{settings.API_V1_STR}/deployments")
        == f"{settings.DEPLOYMENT_API_URL}{settings.API_V1_STR}/deployments"
    )

    assert client.get_path() == f"{settings.DEPLOYMENT_API_URL}"


def test_request_failures():
    client = API(api_url=settings.DEPLOYMENT_API_URL)
    response = client.post("non/existent/path")
    with pytest.raises(requests.exceptions.RequestException):
        response.raise_for_status()

    response = client.get("non/existent/path")
    with pytest.raises(requests.exceptions.RequestException):
        response.raise_for_status()

    response = client.patch("non/existent/path")
    with pytest.raises(requests.exceptions.RequestException):
        response.raise_for_status()

    response = client.put("non/existent/path")
    with pytest.raises(requests.exceptions.RequestException):
        response.raise_for_status()

    response = client.delete("non/existent/path")
    with pytest.raises(requests.exceptions.RequestException):
        response.raise_for_status()


def test_interpret_response():
    client = API(api_url=settings.DEPLOYMENT_API_URL)
    response = client.post("non/existent/path")
    with pytest.raises(APIClientException):
        interpret_response(response, model=BaseModel)

    client = API(api_url="https://my-json-server.typicode.com/typicode/demo/posts/1")
    response = client.get()

    class TestModel(BaseModel):
        id: int
        title: str

    interpreted = interpret_response(response, TestModel)
    assert interpreted.dict()["id"] == 1



import os
from typing import Dict

from docker.errors import APIError  # type: ignore
from pydantic import ValidationError

from tktl.core.exceptions.exceptions import (
    MissingDocker,
    NoContentsFoundException,
    UserRepoValidationException,
)
from tktl.core.loggers import LOG
from tktl.core.managers.docker import DockerManager
from tktl.core.managers.project import ProjectManager
from tktl.core.schemas.project import ProjectValidationOutput
from tktl.core.validation.outputs import (
    ConfigFileValidationFailure,
    ProjectValidationFailure,
)


def validate_project_config(path: str):
    try:
        ProjectManager.validate_project_config(path)
    except ValidationError as e:
        validation_output = ProjectValidationOutput(
            title=ConfigFileValidationFailure.title,
            summary=ConfigFileValidationFailure.summary,
            text=ConfigFileValidationFailure.format_step_results(validation_errors=e),
        )
        log_failure(validation_output)
        return
    except (NoContentsFoundException, UserRepoValidationException) as e:
        validation_output = ProjectValidationOutput(
            title=ProjectValidationFailure.title,
            summary=ProjectValidationFailure.summary,
            text=ProjectValidationFailure.format_step_results(validation_errors=e),
        )
        log_failure(validation_output)
        return
    LOG.log("Project scaffolding is valid!", color="green")


def log_failure(validation_output: ProjectValidationOutput):
    LOG.log(f"Project scaffolding is invalid: {validation_output.title}", color="red")
    LOG.log(validation_output.summary, color="red", err=True)
    LOG.log(validation_output.text, color="red", err=True)


def build_image(path: str, nocache: bool = False, secrets: Dict[str, str] = None):
    dm = DockerManager(path)
    LOG.log("Building docker image...")
    abs_path = os.path.abspath(os.path.join(path, ".buildfile"))
    image = dm.build_image(dockerfile=abs_path, nocache=nocache, buildargs=secrets)

    return dm, image


def status_output(status):
    if status["StatusCode"] == 0:
        LOG.log("Success", color="green")
    else:
        LOG.log("Error", color="red")
        exit(1)


def validate_import(path: str, nocache: bool, secrets: Dict[str, str] = None):
    try:
        dm, image = build_image(path, nocache=nocache, secrets=secrets)
        LOG.log("Testing import of tktl...")
        status, _ = dm.test_import(image)
        status_output(status)
        if nocache:
            dm.remove_image(image=image)

    except MissingDocker:
        LOG.log(
            "Couldn't locate docker, please make sure it is installed, or use the DOCKER_HOST environment variable",
            color="red",
        )


def validate_unittest(path: str, nocache: bool, secrets: Dict[str, str] = None):
    try:
        dm, image = build_image(path=path, nocache=nocache, secrets=secrets)
        LOG.log("Running unittests...")
        status, _ = dm.test_unittest(image)
        status_output(status)
        if nocache:
            dm.remove_image(image=image)

    except MissingDocker:
        LOG.log(
            "Couldn't locate docker, please make sure it is installed, or use the DOCKER_HOST environment variable",
            color="red",
        )


def validate_integration(
    path: str, nocache: bool, timeout: int, retries: int, secrets: Dict[str, str] = None
):
    try:
        dm, image = build_image(path, nocache=nocache, secrets=secrets)
        LOG.log("Waiting for service to start...")
        (
            rest_response,
            grpc_response,
            arrow_container,
            rest_container,
        ) = dm.run_and_check_health(
            image,
            kill_on_success=True,
            auth_enabled=False,
            timeout=timeout,
            retries=retries,
        )
        if _validate_container_response(
            rest_response=rest_response, grpc_response=grpc_response
        ):
            LOG.log("Success", color="green")
        else:
            LOG.log(
                "Unable to run container. See stack trace for more info", color="red"
            )
            exit(1)
        if nocache:
            dm.remove_image(image=image)

    except MissingDocker:
        LOG.log(
            "Couldn't locate docker, please make sure it is installed, or use the DOCKER_HOST environment variable",
            color="red",
        )


def validate_profiling(
    path: str, nocache: bool, timeout: int, retries: int, secrets: Dict[str, str] = None
):
    try:
        dm, image = build_image(path, nocache=nocache, secrets=secrets)
        LOG.log("Testing profiling...")
        LOG.log("Initiating service...")
        (
            rest_response,
            grpc_response,
            arrow_container,
            rest_container,
        ) = dm.run_and_check_health(
            image,
            kill_on_success=False,
            auth_enabled=False,
            timeout=timeout,
            retries=retries,
        )
        success = _validate_container_response(
            rest_response=rest_response, grpc_response=grpc_response
        )
        if not success:
            LOG.log(
                "Failed to run service container, ensure service can run with `tktl validate integration`",
                color="red",
            )
            exit(1)
        try:
            LOG.log("Initiating remote profiling...")
            status, container = dm.run_profiling_container()
            status_output(status)
            if nocache:
                dm.remove_image(image=container.image.id)
        finally:
            try:
                arrow_container.kill()
                rest_container.kill()
                if nocache:
                    dm.remove_image(image=rest_container.image.id)
                    dm.remove_image(image=arrow_container.image.id)
            except APIError:
                pass

    except MissingDocker:
        LOG.log(
            "Couldn't locate docker, please make sure it is installed, or use the DOCKER_HOST environment variable",
            color="red",
        )


def _validate_rest_response(rest_response):
    if rest_response is None:
        LOG.log("Could not access REST endpoint", color="red")
    elif rest_response.status_code != 204:
        LOG.log(f"Response status code {rest_response.status_code}", color="red")
    else:
        return True
    return False


def _validate_grpc_response(grpc_response):
    if grpc_response is None:
        LOG.log("Could not access gRPC endpoint", color="red")
    else:
        return True
    return False


def _validate_container_response(rest_response, grpc_response):
    return _validate_rest_response(
        rest_response=rest_response
    ) and _validate_grpc_response(grpc_response=grpc_response)


def validate_all(
    path: str, nocache: bool, timeout: int, retries: int, secrets: Dict[str, str] = None
):
    validate_project_config(path)
    validate_import(path, nocache=nocache, secrets=secrets)
    validate_unittest(path, nocache=False, secrets=secrets)
    validate_integration(
        path, nocache=False, timeout=timeout, retries=retries, secrets=secrets
    )
    validate_profiling(
        path, nocache=False, timeout=timeout, retries=retries, secrets=secrets
    )



from typing import Any, Dict, List, Sequence, Tuple, Type, Union

import numpy  # type: ignore
import pandas  # type: ignore
from pydantic import BaseModel, parse_obj_as
from pydantic.main import ModelMetaclass

from tktl.core.serializers.base import CustomDeserializingModelT, ObjectSerializer
from tktl.core.t import RestSchemaTypes


class DataFrameSerializer(ObjectSerializer):
    @classmethod
    def deserialize(cls, value: CustomDeserializingModelT) -> pandas.DataFrame:
        return value.deserialize()

    @classmethod
    def serialize(
        cls, value: pandas.DataFrame, output_model: Type[BaseModel] = None
    ) -> List[Tuple[str, BaseModel]]:
        assert output_model
        with_index = [
            {**rec, **{"index": value.index[i]}}
            for i, rec in enumerate(value.to_dict(orient="records"))
        ]
        return [p for p in output_model.parse_obj(with_index)]


class SeriesSerializer(ObjectSerializer):
    @classmethod
    def deserialize(cls, value: CustomDeserializingModelT) -> pandas.DataFrame:
        return value.deserialize()

    @classmethod
    def serialize(
        cls, value: pandas.Series, output_model: Type[BaseModel] = None
    ) -> BaseModel:
        assert output_model
        name = _get_prop_from_series_schema(output_model)
        return output_model(**{name: value.tolist()})


class ArraySerializer(ObjectSerializer):
    @classmethod
    def deserialize(cls, value: CustomDeserializingModelT) -> pandas.DataFrame:
        return value.deserialize()

    @classmethod
    def serialize(
        cls, value: numpy.ndarray, output_model: Type[BaseModel] = None
    ) -> BaseModel:
        assert output_model
        if not isinstance(output_model, ModelMetaclass):
            return parse_obj_as(output_model, value.tolist())
        elif output_model.__name__.startswith(RestSchemaTypes.FLAT_ARRAY.value):
            return output_model(value=value.tolist())
        elif output_model.__name__.startswith(RestSchemaTypes.ARRAY.value):
            return output_model.parse_obj(value.tolist()).__root__
        elif output_model.__name__.startswith(RestSchemaTypes.SERIES.value):
            out_name = _get_prop_from_series_schema(output_model)
        else:
            out_name = "value"
        return output_model(**{out_name: value.tolist()})


class SequenceSerializer(ObjectSerializer):
    @classmethod
    def deserialize(cls, value: CustomDeserializingModelT) -> Any:
        return value

    @classmethod
    def serialize(
        cls, value: Union[Sequence, Dict], output_model: Type[BaseModel] = None
    ) -> Union[Dict, List[Dict]]:
        return [v for v in value]


class PassThroughSerializer(ObjectSerializer):
    @classmethod
    def deserialize(cls, value: CustomDeserializingModelT) -> Any:
        return value

    @classmethod
    def serialize(
        cls, value: BaseModel, output_model: Type[BaseModel] = None
    ) -> BaseModel:
        return value


def _get_prop_from_series_schema(schema):
    return list(schema.schema()["properties"].keys())[0]
